<?php
/*
 * MTF Classes 1.0
 * Definitions of classes used to parse MTF files.
 * Copyright (c) 2003 by Michal Nazarewicz (mina86@tlen.pl)
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */


if ( !defined('IN_PHPBB') )
{
	die("Hacking attempt");
}



class MTFList {
	function MTFList() {
		$this->_list = array();
	}

	function count() {
		return count($this->_list);
	}

	function &get($index) {
		return isset($this->_list[$index])?$this->_list[$index]:null;
	}

	function add(&$item) {
		$this->_list[] = &$item;
	}

	function set($index, &$item) {
		if (!$this->_check_index($index)) {
			return false;
		}
		$this->_list[$index] = &$item;
	}

	function indexOf(&$item, $start = 0) {
		if (!$this->_check_index($start)) {
			return -1;
		}
		$list = &$this->_list;
		$count = count($list);
		for (; $start<$count; ++$start) {
			if ($list[$start]===$item) {
				return $start;
			}
		}
		return -1;
	}

	function lastIndexOf(&$item, $start = null) {
		$list = &$this->_list;
		if ($start===null) {
			$start = count($list)-1;
		}
		if (!$this->_check_index($start)) {
			return -1;
		}
		for (; $start>-1; --$start) {
			if ($list[$start]===$item) {
				return $start;
			}
		}
		return -1;
	}

	function remove($index) {
		if (!$this->_check_index($index)) {
			return false;
		}
		unset($this->_list[$index]);
		$this->_list = array_vars($this->_list);
	}


	function _check_index($index) {
		return intval($index)==$index &&
			$index>=0 && $index<count($this->_list);
	}

	var $_list;
}




define('MTF_PART_MAIN', 0);
define('MTF_PART_PROTOCOL', 1);
define('MTF_PART_PLUGIN', 2);


class MTFPart {
	function MTFPart() {
		$this->_locale = null;
		$this->_language = null;
		$this->_lang_code = null;

		$this->_type = null;
		$this->_name = null;

		$this->_ctime = null;
		$this->_mtime = null;

		$this->_version = null;

		$this->sections = &new MTFList();
		$this->authors = &new MTFList();
		$this->coders = &new MTFList();

		$this->_get_line_type = null;
		$this->_get_line_data = null;
		$this->_get_line_data2 = null;
	}


	function getLocale() {
		return $this->_locale;
	}

	function setLocale($locale = null) {
		$this->_locale = $locale;
	}


	function getLanguage() {
		return $this->_language;
	}

	function setLanguage($language = null) {
		$this->_language = $language;
	}


	function getLangCode() {
		return $this->_lang_code;
	}

	function setLangCode($lang_code = null) {
		$this->_lang_code = $lang_code;
	}



	function getType() {
		return $this->_type;
	}

	function setType($type) {
		$this->_type = $type;
	}


	function getName() {
		if ($this->_type==MTF_PART_MAIN) {
			return 'Main Part';
		} else {
			return $this->_name;
		}
	}

	function setName($name) {
		$this->_name = $name;
	}



	function getCTime() {
		return $this->_ctime;
	}

	function setCTime($ctime = null) {
		$this->_ctime = $time;
	}


	function getMTime() {
		return $this->_mtime;
	}

	function setMTime($time = null) {
		$this->_mtime = $time;
	}


	function getVersion() {
		return $this->_version;
	}

	function setVersion($version) {
		$this->_version = $version;
	}


	function &parseString($data, &$authors_set) {
		$this->_get_line_type = 's';
		$this->_get_line_data = explode("\n", $data);
		$this->_get_line_data2 = 0;

		$ret = &$this->_parse($authors_set);

		$this->_get_line_type = null;
		$this->_get_line_data = null;
		$this->_get_line_data2 = null;

		return $ret;
	}

	function &parseFile($filename, &$authors_set) {
		$fp = fopen($filename, 'r');
		if (!$fp) {
			return 'Could not open file.';
		}

		$this->_get_line_type = 'f';
		$this->_get_line_data = $fp;
		$this->_get_line_data2 = null;

		$ret = &$this->_parse($authors_set);

		fclose($fp);
		$this->_get_line_type = null;
		$this->_get_line_data = null;

		return $ret;
	}


	function _getLine() {
		if ($this->_get_line_type == null || $this->_get_line_data == null) {
			return null;
		}

		switch ($this->_get_line_type) {
			case 'f':
				if (!feof($this->_get_line_data)) {
					return fgets($this->_get_line_data, 4096);
				}
				break;

			case 's':
				if (isset($this->_get_line_data[$this->_get_line_data2])) {
					return $this->_get_line_data[$this->_get_line_data2++];
				}
				break;
		}

		$this->_get_line_type = null;
		$this->_get_line_data = null;
		$this->_get_line_data2 = null;
		return null;
	}


	function _parse(&$authors_set) {
		// Init
		$section = null;
		$string = null;
		$defined = array();
		$errors = false;


		$i = 1;
		for ($i = 1; ($line = &$this->_getLine())!==null; ++$i) {
			$line = rtrim($line);

			// Skip comments and empty lines
			if (strlen($line)==0 ||
				(substr($line, 0, 1)==';' && substr($line, 1, 1)!='#')) {
				continue;
			}

			// Directive
			if (preg_match('/^;#([a-z0-9A-Z\-]*)(\s+.*)?$/', $line, $matches)) {
				$directive = &$matches[1];
				$line = trim($matches[2]);

				// Description
				if ($directive=='') {
					continue;
				}

				// Check if directive is sued after section start
				if ($directive!='section' && $section!==null) {
					echo("(!!) Directive not allowed after section start (line $i)\n");
					$errors = true;
					continue;
				}


				// Check if it wasn't defined already
				if (substr($directive, -2)!='by' && $directive!='section') {
					if (isset($defined[$directive])) {
						echo("(!!) #$directive already defined at line {$defined[$directive]} (line $i)\n");
						$errors = true;
						continue;
					}
					$defined[$directive] = $i;
				}


				// Parse
				switch ($directive) {
					// Language
					case 'language':
						if (!preg_match('/^([0-9]+\s+)?([^(]+)(\((.+)\))\s*$/',
								$line, $matches)) {
							echo("(!!) Syntax error in #lanvguage directive (line $i)\n");
							$errors = true;
							break;
						}
						if (trim($matches[1])) {
							$this->_locale = trim($matches[1]);
						}
						$this->_language = trim($matches[2]);
						if (trim($matches[4])) {
							$this->_lang_code = trim($matches[4]);
						}
						break;


					//
					case 'type':
						if ($line!='main' && $line!='plugin' && $line!='protocol') {
							echo("(!!) Wrong type ($line) (line $i)\n");
							$errors = true;
							break;
						}

						if ($line=='main' && isset($defined['name'])) {
							echo("(==) Main cannot have #name directive (line $i)\n");
						}

						if ($line=='main') {
							$this->_type = MTF_PART_MAIN;
						} elseif ($line=='protocol') {
							$this->_type = MTF_PART_PROTOCOL;
						} else {
							$this->_type = MTF_PART_PLUGIN;
						}
						break;


					// Name
					case 'name':
						if ($this->_type===MTF_PART_MAIN) {
							echo("(==) Main cannot have #name directive (line $i)\n");
						}
						if ($line=='') {
							echo("(!!) Missing argument in #name directive (line $i)\n");
							$errors = true;
						}
						$this->_name = $line;
						break;


					// Version
					case 'version':
						if ($line=='') {
							echo("(==) Missing argument in #version directive (line $i)\n");
						}
						$this->_version = $line;
						break;


					// Release
					case 'release':
						if ($line=='') {
							echo("(==) Missing argument in #release directive (line $i)\n");
							break;
						}

						$line = explode(',', $line);
						if (count($line)>2) {
							echo("(==) Too many parameters for #release directive (line $i)\n");
						}

						// CTime
						if (trim($line[0])) {
							$date = mtf_date2time(trim($line[0]));
							if ($date==0) {
								echo("(!!) Wrong date format in #release directive (line $i)\n");
								$errors = true;
							} else {
								$this->_ctime = $date;
							}
						}
						if (count($line)==1) {
							break;
						}

						// MTime
						if (!trim($line[1])) {
							echo("(==) Missing parameter after ',' in #release directive (line $i)\n");
							break;
						}
						$date = mtf_date2time(trim($line[1]));
						if ($date==0) {
							echo("(!!) Wrong date format in #release directive (line $i)\n");
							$errors = true;
							break;
						}
						$this->_mtime = $date;
						break;


					// Coder & translator
					case 'by':
					case 'translated-by':
						if (!$line) {
							echo("(==) Missing parameter in #$directive directive (line $i)\n");
							break;
						}

						$line = explode(';', $line);
						for ($j = 0; $j<count($line); ++$j) {
							if (!preg_match('/^(.+?)\s*(\((.+)\))?\s*$/',
								trim($line[$j]), $matches)) {
								echo("(!!) Syntax error in #$directive directive (line $i)\n");
								$errors = true;
								break;
							}

							$author = &$authors_set->get(trim($matches[1]), true);
							if ($directive=='by') {
								$this->coders->add($author);
								$author->coded->add($this);
							} else {
								$this->authors->add($author);
								$author->translated->add($this);
							}

							if (count($matches)>2) {
								$line = explode(',', $matches[3]);
								for ($k = 0; $k<count($line); ++$k) {
									$info = trim($line[$k]);
									if ($info=='') {
										echo("(==) Empty info in #$directive directive (line $i)\n");
									} elseif ($author->info->indexOf($info)==-1) {
										$author->info->add($info);
									}
								}
							}
						}
						break;


					// Section
					case 'section':
						if ($string!==null) {
							echo("(--) There was string without translation (before line $i)\n");
							$section->addString($string);
							$string = null;
						}

						unset($section);
						$section = &new MTFSection($line);
						$this->sections->add($section);
						break;

					// Unknown direcitve
					default:
						echo("(==) Unknown directive #$directive (line $i)\n");
						break;
				}

			// String start
			} elseif  (substr($line, 0, 1)=='[' && substr($line, -1)==']') {
				if ($section===null) {
					echo("(!!) #section required before [string] (line $i)\n");
					$errors = true;
					break;
				}
				if ($string!==null) {
					echo("(--) There was string without translation (before line $i)\n");
					$section->addString($string);
				}
				$string = substr($line, 1, -1);

			// Translation (no string to translate)
			} elseif ($string===null) {
				echo("(==) There was no string to be translated (line $i)\n");

			// Translation
			} else {
				$section->addString($string, $line);
				$string = null;
			}
		}

		return !$errors;
	}



	var $_language;
	var $_locale;
	var $_lang_code;

	var $_type;
	var $_name;

	var $_ctime;
	var $_mtime;

	var $_version;

	var $sections;
	var $authors;
	var $coders;

	var $_get_line_type;
	var $_get_line_data;
	var $_get_line_data2;
}





class MTFSection {
	function MTFSection($name) {
		$this->_name = $name;
		$this->_strings = array();
	}


	function getName() {
		return $this->_name;
	}

	function setName($name) {
		$this->_name = $name;
	}



	function countStrings() {
		return count($this->_strings);
	}

	function getString($str) {
		return isset($this->_strings[$str])?$this->_strings[$str]:false;
	}

	function addString($str, $translation = null) {
		$this->_strings[$str] = $translation;
	}

	function removeString($str) {
		if (isset($this->_strings[$str])) {
			unset($this->_strings[$str]);
		}
	}

	function resetStrings() {
		reset($this->_strings);
	}

	function &eachString() {
		return each($this->_strings);
	}



	var $_name;
	var $_strings;
}





class MTFAuthor {
	function MTFAuthor($name) {
		$this->_name = $name;
		$this->info = &new MTFList();
		$this->translated = &new MTFList();
		$this->coded = &new MTFList();
	}


	function getName() {
		return $this->_name;
	}

	function setName($name) {
		$this->_name = $name;
	}


	var $_name;
	var $info;
	var $translated;
	var $coded;
}



class MTFAuthorsSet {
	function MTFAuthorsSet() {
		$this->authors = array();
	}

	function &get($name, $create = false) {
		if (isset($this->authors[$name])) {
			return $this->authors[$name];
		} elseif (!$create) {
			return null;
		} else {
			$author = &new MTFAuthor($name);
			$this->authors[$name] = &$author;
			return $author;
		}
	}

	function set($name, &$author) {
		$this->authors[$name] = &$author;
	}

	function create($name, &$author) {
		if (isset($this->authors[$name])) {
			return false;
		}
		$this->authors[$name] = &$author;
		return true;
	}

	function remove($name) {
		if (isset($this->authors[$name])) {
			return false;
		}
		unset($this->authors[$name]);
		return true;
	}


	var $authors;
}



class MTFPartsSet {
	function MTFPartsSet() {
		$this->_main = null;
		$this->protocols = &new MTFList();
		$this->plugins = &new MTFList();
		$this->authors = &new MTFAuthorsSet();
	}

	function countAllPlugins() {
		return $this->_protocols->count() + $this->_plugins->count();
	}


	function &getMainPart() {
		return $this->_main;
	}

	function setMainPart(&$main) {
		$this->_main = &$main;
	}


	function parseStrings($array) {
		$warnings = '';
		for ($i = 0; $i<count($array); ++$i) {
			$part = &new MTFPart();
			$warnings .= "Parsing string #$i:\n";
			$warnings .= $part->parseString($array[$i], $this->authors);
			switch ($part->getType()) {
				case MTF_PART_PROTOCOL:
					$this->protocols->add($part);
					break;
				case MTF_PART_PLUGIN:
					$this->plugins->add($part);
					break;
				case MTF_PART_MAIN:
					if ($this->_main!==null) {
					}
					$this->_main = &$part;
					break;
				default:
					$warnings .= "Unknown type (" . $part->getType() . ")\n";
					break;
			}
			$warnings .= "\n";
		}
	}

	function parseFiles($array) {
		$warnings = '';
		for ($i = 0; $i<count($array); ++$i) {
			$part = &new MTFPart();
			$warnings .= "Parsing file {$arry[$i]}:\n";
			$warnings .= $part->parseFile($array[$i], $this->authors);
			switch ($part->getType()) {
				case MTF_PART_PROTOCOL:
					$this->protocols->add($part);
					break;
				case MTF_PART_PLUGIN:
					$this->plugins->add($part);
					break;
				case MTF_PART_MAIN:
					if ($this->_main!==null) {
					}
					$this->_main = &$part;
					break;
				default:
					$warnings .= "Unknown type (" . $part->getType() . ")\n";
					break;
			}
			$warnings .= "\n";
		}
	}


	var $_main;
	var $protocols;
	var $plugins;
	var $authors;
}






define('MTF_PRINTER_NO_COMMENT', 1);
define('MTF_PRINTER_COMMENT_DUP', 2);
define('MTF_PRINTER_COMMENT_NOT_TRANSLATED', 4);

class MTFPrinter {
	function MTFPrinter($flags = 0, $thanks = null, $notes = null) {
		$this->reset($flags, $thanks, $notes);
	}

	function reset($flags = 0, $thanks = null, $notes = null) {
		$this->_flags = $flgas;
		$this->_thanks = $thanks;
		$this->_notes = $notes;
		$this->_strings = array();
	}


	function printAll(&$parts) {
		$main = &$parts->getMainPart();
		if ($main===null) {
			echo("; ERROR: No main part.");
			return false;
		}
		return
			$this->printHeaders($parts) &&
			$this->printNotes() &&
			$this->printPart($main) &&
			$this->printParts($parts->protocols) &&
			$this->printParts($parts->plugins);
	}


	function printHeaders(&$parts) {
		echo("Miranda Language Pack Version 1");
		$main = &$parts->getMainPart();
		if ($main===null) {
			echo("\n; ERROR: No main part.");
			return false;
		}

		if ($main->getLanguage()!==null) {
			echo("\nLanguage: " . $main->getLanguage());
			if ($main->getLangCode()!==null) {
				echo(" (" . $main->getLangCode() . ")");
			}
		}

		if ($main->getLocale()!==null) {
			echo("\nLocale: " . substr('0000' . $main->getLocale(), -4));
		}

		if ($main->getVersion()!==null) {
			echo("\nLast-Modified-Using: Miranda IM " . $main->getVersion());
		}

		if ($this->_thanks!==null) {
			echo("\nThanks: {$this->_thanks}");
		}

		return
			$this->printAuthorsHeader($parts->authors) &&
			$this->printPluginsHeader($parts->protocols) &&
			$this->printPluginsHeader($parts->plugins);
	}


	function printAuthorsHeader(&$aset) {
		$authors = &$aset->authors;
		$len = 300;
		$first = true;
		foreach ($authors as $key => $object) {
			if ($object->translated->count()!=0) {
				if ($len>200) {
					echo("\nAuthors: ");
					$first = true;
				}
				$name = &$object->getName();
				$len += strlen($name);
				echo(($first?'':', ') . $name);
				$first = false;
			}
		}
		return true;
	}


	function printPluginsHeader(&$plugins) {
		$len = 300;
		$first = true;
		for ($i = 0; $i<$plugins->count(); ++$i) {
			if ($len>200) {
				echo("\nPlugins-included: ");
				$first = true;
			}
			$p = &$plugins->get($i);
			$name = &$p->getName();
			$len += strlen($name);
			echo(($first?'':', ') . $name);
			$first = false;
		}
		return true;
	}

	function printNotes() {
		if ($this->_notes!==null) {
			echo( "\n\n; " . preg_replace("/\n[ \t]*/", "\n; ", trim($this->_notes)) . "\n");
		}
		return true;
	}


	function printParts(&$parts) {
		for ($i = 0; $i<$parts->count(); ++$i) {
			if (!$this->printPart($parts->get($i))) {
				return false;
			}
		}
		return true;
	}


	function printPart(&$part) {
		echo("\n");
		if ($this->_flags & MTF_PRINT_NO_COMMENT==0) {
			echo("\n\n");
			if ($part->getType()==MTF_PART_MAIN) {
				echo("; ***** Main translation *****\n");
			} else {
				echo("; ***** " . $part->getName() . "*****\n");
			}
			if ($part->getVersion()!==null) {
				echo("; " . $part->getVersion() . "\n");
			}
		}

		$sections = &$part->sections;
		for ($i = 0; $i<$sections->count(); ++$i) {
			if (!$this->printSection($sections->get($i))) {
				return false;
			}
		}
		return true;
	}


	function printSection(&$section) {
		if ($this->_flags & MTF_PRINT_NO_COMMENT==0) {
			echo("\n");
			echo("; * " . $section->getName() . "\n");
		}

		$section->resetStrings();
		while (($string = $section->eachString())) {
			list($string, $trans) = $string;
			if ($trans===null) {
				if ($this->_flags & MTF_PRINT_COMMENT_NOT_TRANSLATED!=0) {
					echo(";[$string]\n; Not translated\n");
				}
				continue;
			}

			if (isset($this->_strings[$string])) {
				if ($this->_flags & MTF_PRINT_COMMENT_DUP!=0) {
					echo(";[$string]\n;$trans\n; Duplication\n");
				}
				continue;
			}

			echo("[$string]\n$trans\n");
		}
		return true;
	}


	var $_flags;
	var $_thanks;
	var $_strings;
}









function mtf_date2time($date) {
	$len = strlen($date);
	if ($len<8 || $len>14 || $len&1==1 || preg_match('/[^0-9]/', $date)) {
		return 0;
	}

	$year =         intval(substr($date,  0, 4))  ;
	$mon  =         intval(substr($date,  4, 2))  ;
	$day  =         intval(substr($date,  6, 2))  ;
	$hour = $len> 8?intval(substr($date,  8, 2)):0;
	$min  = $len>10?intval(substr($date, 10, 2)):0;
	$sec  = $len>12?intval(substr($date, 12, 2)):0;

	return mktime($hour, $min, $sec, $mon, $day, $year);
}

function mtf_time2date($time) {
	$str = date('YmdHis');
	for ($i = 0; substr($str, -2)=='00' && $i<3; ++$i) {
		$str = substr($str, 0, -2);
	}
	return $str;
}
?>
